class WebDriverError(Exception):
    """
    Base class for more specific WebDriverErrors
    """

    pass


class TabNotOpenException(WebDriverError):
    """
    Throw this when a tab is expected to be open(ed), but it does not
    """

    pass



import requests


class SNAPI:
    """
    Interface to the Selenium Node API.
    """

    @staticmethod
    def start(ip_address: str):
        """
        Starts a Selenium Server instance at an available port, returning the port number.
        :return:
        """
        return (
            requests.post(f"http://{ip_address}:8000/port", timeout=60)
            .json()
            .get("port", "")
        )

    @staticmethod
    def status(ip_address: str, port: str):
        """
        Checks whether or not the Selenium Server instance at the given port has started.
        :return:
        """
        return (
            requests.get(f"http://{ip_address}:8000/port/{port}", timeout=60)
            .json()
            .get("status", "")
            == "up"
        )

    @staticmethod
    def public_ip(ip_address: str):
        """
        Returns the public-facing IP address of the server.
        :return:
        """
        return (
            requests.get(f"http://{ip_address}:8000/public-ip", timeout=60)
            .json()
            .get("ip_address", "")
        )

    @staticmethod
    def kill_watchdog(ip_address: str):
        """
        Stops the watchdog service from running on the server.
        :return:
        """
        return (
            requests.post(
                f"http://{ip_address}:8000/kill-watchdog", timeout=60
            ).status_code
            == 200
        )

    @staticmethod
    def stop(ip_address: str, port: str, session_dir: str = None):
        """
        Stops the Selenium Server instance running at the specified port.
        Wipes the directory specified on the server, if any.
        :return:
        """
        return (
            requests.delete(
                f"http://{ip_address}:8000/port/{port}",
                data={"session_directory": session_dir} if session_dir else {},
                timeout=60,
            ).status_code
            == 200
        )

    @staticmethod
    def clear_session(ip_address: str, session_dir: str):
        return (
            requests.delete(
                f"http://{ip_address}:8000/session",
                data={"session_directory": session_dir},
                timeout=60,
            ).status_code
            == 200
        )



import time
from deprecation import deprecated
from selenium.webdriver.remote.webdriver import WebDriver as SeleniumRemoteWebDriver
from dill.source import getsource
from pytest_drivings.exceptions import TabNotOpenException
from seleniumrequests.request import RequestMixin


class BaseWebDriver(SeleniumRemoteWebDriver, RequestMixin):
    def open_blank_tab(self):
        """
        Open a blank tab in the background, stay on the current tab.
        Keeps the webdriver focused on the tab that was focused previously.
        :return: the window handle of the tab that was just opened
        """
        prev_handles = set(self.window_handles)
        prev_handle = self.current_window_handle
        new_tab_script = f"window.open('about:blank', 'new_tab_{len(prev_handles)}')"
        self.execute_script(new_tab_script)
        current_handles = set(self.window_handles)
        new_handles = current_handles.difference(prev_handles)
        if len(new_handles) < 1:
            raise TabNotOpenException(
                "open_blank_tab did not open a blank tab "
                "after executing the following JavaScript:\n"
                f"{new_tab_script}"
            )
        self.switch_to.window(prev_handle)
        return new_handles[0]

    def load_page_in_tab(self, window_handle, page_class, *page_args, **page_kwargs):
        """
        Creates an instance of the given page class in the tab represented by the given window handle.
        Keeps the webdriver focused on the tab that was focused previously.

        Passing a webdriver into this function for page construction is OPTIONAL.
        If no webdriver is supplied, this instance will be injected into the constructor.

        :param window_handle: unique ID for the tab you wish to load a new page in.
        :param page_class: the class of the page you want to instantiate
        :param page_args: positional arguments to the page constructor
        :param page_kwargs: keyword arguments to the page constructor
        :return: ( newly constructed instance of your page, window handle of the new tab )

        Example Usage:
            - Load an instance of the LoginPage in a given window, switch to the new handle:
                new_page, handle = webdriver.load_page_in_tab(handle, LoginPage, next_page=HomePage, pre_loaded=False)
                webdriver.switch_to.window(handle)
        """
        prev_handle = self.current_window_handle
        self.switch_to.window(window_handle)
        # decide if we need to supply a webdriver, or if one is supplied
        need_wd = any(
            [
                isinstance(arg, BaseWebDriver)
                for arg in page_args + tuple(page_kwargs.values())
            ]
        )
        # return an instance of the page
        page_instance = (
            page_class(self, *page_args, **page_kwargs)
            if need_wd
            else page_class(*page_args, **page_kwargs)
        )
        self.switch_to.window(prev_handle)
        return page_instance, window_handle

    def load_page_in_new_tab(self, page_class, *page_args, **page_kwargs):
        """
        Creates an instance of the given page class in a NEW tab.
        Keeps the webdriver focused on the tab that was focused previously.

        Passing a webdriver into this function for page construction is OPTIONAL.
        If no webdriver is supplied, this instance will be injected into the constructor.

        :param page_class: the class of the page you want to instantiate
        :param page_args: positional arguments to the page constructor
        :param page_kwargs: keyword arguments to the page constructor
        :return: ( newly constructed instance of your page, window handle of the new tab )

        Example usage:
            - Open a new tab and force the webdriver to load a new instance of the LoginPage with custom kwargs:
                new_page, new_window = webdriver.load_page_in_new_tab(LoginPage, next_page=HomePage, pre_loaded=False)
                webdriver.switch_to.window(new_window)
        """
        new_handle = self.open_blank_tab()
        return self.load_page_in_tab(new_handle, page_class, *page_args, **page_kwargs)

    def open_new_tab_via_action(self, action, *action_args, **action_kwargs):
        """
        Execute a given action.
        Assert that the action opened a new tab.
        Direct the webdriver to focus on the new tab.

        :param action: the action to be executed
        :param action_args: positional arguments for the action
        :param action_kwargs: keyword arguments for the action
        :return: ( return value of the supplied action, window handle of the newly opened tab )

        Example usage:
            - Open new tab via button press, create an instance of the page that you opened:
                webdriver.open_new_tab_via_action(lambda: my_button.click(webdriver))
                new_page = MyPage(webdriver, pre_loaded=True)

            - Open new tab via function call, passing in args and kwargs:
                return_value, new_handle = webdriver.open_new_tab_via_action(my_function, arg1, arg2,
                                                                             kwarg1=val1, kwarg2=val2)
                new_page = MyPage(web_driver, pre_loaded=True)
        """
        prev_handles = set(self.window_handles)
        action_ret_val = action(*action_args, **action_kwargs)
        current_handles = set(self.window_handles)
        new_handles = current_handles.difference(prev_handles)
        if len(new_handles) < 1:
            raise TabNotOpenException(
                f"The action provided to open_new_tab_via_action did not open a new tab. \n"
                f"{self._get_action_debug_string(action, *action_args, **action_kwargs)}"
            )
        self.switch_to.window(list(new_handles)[0])
        return action_ret_val, list(new_handles)[0]

    @staticmethod
    def _get_action_debug_string(action, *action_args, **action_kwargs):
        result = f"Action: \n {getsource(action)}\n"
        if len(action_args) > 0:
            result += f"Provided args: {action_args} \n"
        if len(action_kwargs) > 0:
            result += f"Provided kwargs: {action_kwargs} \n"
        return result

    @deprecated(
        deprecated_in="1.0.1",
        details="Use `load_page_in_tab/load_page_in_new_tab` instead.",
    )
    def wait_for_new_tab_to_open(self, timeout=10):
        tries = 0
        while len(self.window_handles) == 1:
            if tries > timeout:
                raise TimeoutError("Timeout waiting for the new tab to open")
            time.sleep(1)
            tries += 1

    @deprecated(
        deprecated_in="1.0.1",
        details="Use `load_page_in_tab/load_page_in_new_tab` instead.",
    )
    def wait_for_tab_to_load(self, page, timeout=10):
        tries = 0
        while "about:blank" in self.current_url:
            if tries > timeout:
                raise TimeoutError("Timeout waiting for the tab to load.")
            time.sleep(1)
            tries += 1

    @deprecated(
        deprecated_in="1.0.1", details="Use webdriver.switch_to.window(window_handle)"
    )
    def switch_to_tab(self, handle_index: int):
        self.switch_to.window(self.window_handles[handle_index])

    @deprecated(
        deprecated_in="1.0.1",
        details="Use webdriver.current_window_handle "
        "to get a unique string that identifies your current window",
    )
    def get_current_tab_index(self):
        """
        Function for getting the index of the current tab
        """
        current_handle = self.current_window_handle
        all_handles = self.window_handles
        return all_handles.index(current_handle)

    @deprecated(
        deprecated_in="1.0.1",
        details="Use: webdriver.switch_to.window(window_handle) and then webdriver.close()",
    )
    def close_tab(self, handle_index: int):
        self.switch_to.window(self.window_handles[handle_index])
        self.close()
        if len(self.window_handles) > 0:
            self.switch_to.window(self.window_handles[0])



from selenium.webdriver.chrome.webdriver import WebDriver
from pytest_drivings.base_webdriver import BaseWebDriver


class LocalWebDriver(WebDriver, BaseWebDriver):
    """
    Create local webdrivers using this class to gain access to webdriver helper functions.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_drivings']

package_data = \
{'': ['*']}

install_requires = \
['black>=19.3b0,<20.0',
 'deprecation>=2.0.7,<3.0.0',
 'mypy>=0.720,<0.721',
 'paramiko>=2.7.1,<3.0.0',
 'psutil>=5.6.7,<6.0.0',
 'pytest-xdist>=1.31.0,<2.0.0',
 'selenium-requests>=1.3,<2.0',
 'selenium>=3.141.0,<4.0.0']

entry_points = \
{'pytest11': ['pytest_drivings = pytest_drivings.webdriver']}

setup_kwargs = {
    'name': 'pytest-drivings',
    'version': '1.0.4',
    'description': 'Tool to allow webdriver automation to be ran locally or remotely',
    'long_description': None,
    'author': 'Jonah Caruso',
    'author_email': 'jayc035@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



"""
Plugin configuration. Allows number of vCPUs & GBs of RAM allocated
to each Selenium Server to be specified, as well as the number of
browsers to run on each server and the maximum number of servers
which may be created simultaneously.
"""
from collections import namedtuple
import os

Environment = namedtuple(
    "Environment", ["agent_params", "max_vms", "max_drivers_per_vm"]
)

# New environment, designed to run multiple browsers per VM with session isolation.
# Tested rigorously to achieve an ideal load average (~4.5)
MAX_DRIVER_ENV = Environment(
    agent_params={
        "os": "seleniumserver",
        "cpu_count": 2,
        "ram_amount": 5120,
        "os_volume": {"size": 10},
        "agent_type": "",
        "name": "selenium",
        "ttl": 8,
        "region": "nwk",
    },
    max_vms=45,
    max_drivers_per_vm=4,
)


# Stable environment designed to run multiple browsers per VM with session isolation.
# High RAM to hopefully prevent crashes.
# Drivers per VM tuned for stability
STABLE_ENV = Environment(
    agent_params={
        "os": "seleniumserver",
        "cpu_count": 2,
        "ram_amount": 16384,
        "os_volume": {"size": 10},
        "agent_type": "",
        "name": "selenium",
        "ttl": 8,
        "region": "nwk",
    },
    max_vms=80,
    max_drivers_per_vm=6,
)

SAFE_ENV = Environment(
    agent_params={
        "os": "seleniumserver",
        "cpu_count": 2,
        "ram_amount": 16384,
        "os_volume": {"size": 10},
        "agent_type": "",
        "name": "selenium",
        "ttl": 8,
        "region": "nwk",
    },
    max_vms=80,
    max_drivers_per_vm=os.environ.get("browsers_per_server", 2),
)

ACTIVE_ENV = SAFE_ENV


def max_runners():
    """
    Returns the maximum number of Remote WebDriver instances supported by the active configuration.
    """
    return ACTIVE_ENV.max_drivers_per_vm * ACTIVE_ENV.max_vms



import pytest
from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from pytest_drivings.local_webdriver import LocalWebDriver
from pytest_drivings.base_webdriver import BaseWebDriver as RemoteWebDriver
from datetime import datetime
import logging
from atexit import register
import random
import string
import math
import pytest_drivings.settings as settings
import os
import shutil
import json
import time
import traceback
import sys
from pytest_drivings.snapi import SNAPI
import concurrent.futures as futures

# Debug mode flag - Can be set via GitLab environment variables or local environment variables
DEBUG_MODE = os.environ.get("DEBUG_MODE", 0) == 1

download_path_pref = {"download.default_directory": "/tmp/test-results/downloads"}

# Options/capabilities for a remote, headless Chromedriver instance
remote_options = webdriver.ChromeOptions()
remote_options.add_argument("headless")
remote_options.add_argument("disable-gpu")
remote_options.add_argument("enable-automation")
remote_options.add_argument("disable-extensions")
remote_options.add_argument("window-size=1920,1080")
remote_options.add_argument("start-maximized")
remote_options.add_experimental_option("prefs", download_path_pref)
remote_capabilities = remote_options.to_capabilities()
remote_capabilities["browserName"] = "chrome"
remote_capabilities["platform"] = "Linux"

# Options for a local, headed Chromedriver instance
local_options = webdriver.ChromeOptions()
local_options.add_argument("no-sandbox")
local_options.add_argument("window-size=1920,1080")
local_options.add_argument("start-maximized")
local_options.add_experimental_option("prefs", download_path_pref)

# settings to allow the webdriver to capture chrome console AND network logs
google_logging_prefs = {"browser": "ALL", "performance": "INFO"}


def pytest_addoption(parser):
    """
    Add a command line option to switch between using local and remote WebDrivers
    :param parser:
    :return: None
    """
    parser.addoption(
        "--local",
        action="store_true",
        default=False,
        help="Use local WebDriver (good for debugging tests)",
    )
    parser.addoption(
        "--headless",
        action="store_true",
        default=False,
        help="Run tests in headless mode " "(doesn't take up your screen!)",
    )


def selenium_server_port(server_ip):
    """
    Starts the Selenium Server executable on an available port of this worker's  server instance.
    Returns the port number on which the service was started.
    :return:
    """
    logging.info(
        f"Sending a Snapi request to start a Selenium Server instance on server {server_ip}"
    )
    port_number = SNAPI.start(server_ip)
    logging.info(f"Triggered a Selenium Server instance on {server_ip}:{port_number}")
    started = False
    attempts = 0
    while not started and attempts < 3:
        started = SNAPI.status(server_ip, port_number)
        attempts += 1
        if not started:
            time.sleep(attempts ** 2)
    assert (
        started
    ), f"Selenium Server instance at {server_ip}:{port_number} failed to start."
    logging.info(
        f"Confirmed via Snapi that Selenium Server has started at {server_ip}:{port_number}"
    )
    return port_number


def set_selenium_server_public_ip(server_ip):
    """
    This function executes some python on the selenium server, storing its public-facing IP
    as an environment variable, usable in tests.
    :param server_ip:
    :return:
    """
    result = SNAPI.public_ip(server_ip)
    assert len(result) > 0, "Server external IP could not be resolved, is  running?"
    os.environ["server_external_ip"] = result


def create_webdriver(session_dir="/tmp/chrome", headless=False):
    if headless:
        local_options.add_argument("headless")
    # Ensure each webdriver will have a unique session directory
    local_options.add_argument(f"user-data-dir={session_dir}")

    desired_caps = DesiredCapabilities.CHROME.copy()
    desired_caps["goog:loggingPrefs"] = google_logging_prefs

    local_options.add_argument(f"user-data-dir={session_dir}")
    driver = LocalWebDriver(
        "chromedriver", options=local_options, desired_capabilities=desired_caps
    )
    return driver


@pytest.fixture(scope="function", autouse=True)
def web_driver(request):
    """
    Creates and yields a Local or Remote WebDriver instance,
    then destroys the instance after the test function completes.
    :param request: Request instance.
    :return: None
    """
    # Create a new (random) directory for the new WebDriver's session data
    session_handle = "".join(
        [random.choice(string.ascii_letters + string.digits) for _ in range(10)]
    )
    session_dir = f"/tmp/profile-{session_handle}"
    if request.config.getoption("--local"):
        # Create a local Chromedriver instance with a unique session directory
        driver = create_webdriver(
            session_dir=session_dir, headless=request.config.getoption("--headless")
        )
        # Create function to reset driver session
        driver.clear_session = lambda: wipe_directory(session_dir)
    else:
        # Create Remote WebDriver instance using Selenium Server instance
        remote_options.add_argument(f"user-data-dir={session_dir}")
        remote_caps = remote_options.to_capabilities()
        remote_caps["browserName"] = "chrome"
        remote_caps["platform"] = "Linux"
        remote_caps["goog:loggingPrefs"] = google_logging_prefs
        attempts = 0
        driver = None
        while attempts < 3 and not driver:
            try:
                server = get_server(request.config)
                server_port = get_server_port(request.config)
                set_selenium_server_public_ip(server)
                logging.info(
                    f"Attempting to start a RemoteWebDriver on server {server}:{server_port}"
                )
                driver = RemoteWebDriver(
                    f"http://{server}:{server_port}/wd/hub", remote_caps
                )
            except:
                traceback.print_exc(file=sys.stdout)
                attempts += 1
                time.sleep(10 * (attempts ** 2))
        if DEBUG_MODE and not driver:
            # Keep the server alive to investigate the error
            # Kill the watchdog, Old Yeller style
            logging.debug(
                f"Sending Snapi request to stop watchdog so faulty server {server} can be investigated"
            )
            SNAPI.stop(get_server(request.config), server_port)
            # Extend the TTL by 24 hours so we can investigate in the morning
            logging.debug(
                f"Sending  request to extend server TTL so faulty server {server} ({get_server_id(request.config)}) can be investigated"
            )
            ().renew_agent(get_server_id(request.config), 1)
        assert driver, (
            f"Failed to connect to the Selenium Server with IP address {get_server(request.config)} "
            f"after 3 attempts. Is the server online?"
        )
        # Create function to reset driver session
        driver.clear_session = lambda: SNAPI.clear_session(
            get_server(request.config), session_handle
        )
    # Provide webdriver to test session
    yield driver
    # Post-session teardown
    # Release WebDriver instance
    driver.quit()


def wipe_directory(path):
    """
    Recursively deletes all contents of the folder at the given path.
    :param path:
    :return:
    """
    for root, dirs, files in os.walk(path):
        for f in files:
            os.unlink(os.path.join(root, f))
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))


def get_server(config):
    """
    Returns the IP address of the server assigned to this test runner process.
    :param config:
    :return:
    """
    if hasattr(config, "workerinput"):
        return os.environ.get("server")
    else:
        return json.loads(os.environ.get("servers")).get("data")[0].get("ip")


def get_server_id(config):
    """
    Returns the  ID of the server assigned to this test runner process.

    :param config:
    :return:
    """
    if hasattr(config, "workerinput"):
        return os.environ.get("server_id")
    else:
        return json.loads(os.environ.get("servers")).get("data")[0].get("id")


def get_server_port(config):
    """
    Returns the port number of the Selenium Server instance running on
    the server assigned to this test runner process.
    """
    if hasattr(config, "workerinput"):
        return os.environ.get("port")
    else:
        return json.loads(os.environ.get("servers")).get("data")[0].get("port")


def teardown():
    """
    Tears down Selenium Server and WebDriver instances.
    :return: None
    """
    try:
        for server in json.loads(os.environ.get("servers")).get("data"):
            try:
                # Kill Selenium Server app instance running on our server
                logging.info(
                    f"Tearing down server ({server.get('ip')} ({server.get('id')})) via  DELETE request"
                )
                ().delete_agent(server.get("id"))
            except:
                pass
    except:
        pass


def pytest_keyboard_interrupt(excinfo):
    """
    Ensure that WebDriver and Selenium Server instances are torn down after
    a test run is terminated via a keyboard interrupt.
    :param excinfo: Exception info.
    :return: None
    """
    teardown()


def pytest_sessionfinish(session, exitstatus):
    """
    Ensure that WebDriver and Selenium Server instances are torn down after
    a test run is terminated for any reason.
    :param session: Session data.
    :param exitstatus: Exit status.
    :return: None
    """
    if not (
        session.config.getoption("--local") or hasattr(session.config, "workerinput")
    ):
        teardown()


def pytest_xdist_setupnodes(config, specs):
    """
    Assigns Selenium Server instances to Xdist worker processes.
    :param config:
    :param specs:
    :return:
    """
    if not config.getoption("--local"):
        for i in range(len(specs)):
            server_index = i // settings.ACTIVE_ENV.max_drivers_per_vm
            specs[i].env["server"] = (
                json.loads(os.environ.get("servers"))
                .get("data")[server_index]
                .get("ip")
            )
            specs[i].env["server_id"] = (
                json.loads(os.environ.get("servers"))
                .get("data")[server_index]
                .get("id")
            )
            specs[i].env["port"] = str(
                json.loads(os.environ.get("servers"))
                .get("data")[server_index]
                .get("port")
            )
            # setting automation_foundation driver key for parallel usage
            specs[i].env["AF_DRIVER_KEY"] = specs[i].id


def pytest_configure(config):
    """
    Creates the minimum number of Selenium Server instances required to
    support the requested number of workers.
    :param config:
    :return:
    """
    if not (config.getoption("--local") or hasattr(config, "workerinput")):
        workers = config.getoption("numprocesses")
        if not workers:
            workers = 1
        assert workers <= settings.max_runners(), (
            f"The maximum number of runners currently allowed is "
            f"{settings.max_runners()}."
        )
        vms_to_create = math.ceil(workers / settings.ACTIVE_ENV.max_drivers_per_vm)

        def destroy_servers(s):
            for server in s:
                ().delete_agent(server["id"])

        def attempt_server_creation():
            servers = []
            attempts = 3
            while attempts > 0 and len(servers) < vms_to_create:
                servers += create_servers(vms_to_create - len(servers))
                attempts -= 1
            return servers

        # attempt server creation in Norwalk
        logging.info("Attempting to create Selenium Server VMs in Norwalk")
        start_time = datetime.now()
        servers = attempt_server_creation()
        if len(servers) != vms_to_create:
            destroy_servers(servers)
            logging.warning(
                f"Failed to successfully create {vms_to_create} Selenium Servers in Norwalk after three attempts."
            )
            logging.info(f"Attempting to create fallback servers in Rochester...")
            settings.ACTIVE_ENV.agent_params["region"] = "roc"
            start_time = datetime.now()
            servers = attempt_server_creation()
            if len(servers) != vms_to_create:
                destroy_servers(servers)
                pytest.exit(
                    f"Failed to successfully create {vms_to_create} Selenium Servers "
                    f"in both Rochester and Norwalk after three attempts (each). Is  working?"
                )
        logging.info(
            f" created {vms_to_create} Selenium Server VMs in {datetime.now()-start_time}"
        )
        os.environ["servers"] = json.dumps({"data": servers})
        if DEBUG_MODE:
            logging.debug(
                f"Servers to be used in this test run: {', '.join([server['ip']+':'+server['port'] + ' ('+server['id']+')' for server in servers])}"
            )


def create_server():
    """
    Issues a request to  to create a Selenium Server VM and waits for the agent to start.
    :return:
    """
    agent = ().create_agent(CreateAgentPayload(**settings.ACTIVE_ENV.agent_params))
    agent = ().wait_for_agent_to_start(agent.id)
    port = selenium_server_port(agent.ip_address)
    return {"ip": agent.ip_address, "id": agent.id, "port": port}


def create_servers(num_servers):
    """
    Creates the specified number of Selenium Servers.
    Performs server creation concurrently for speed, rate-limits requests for stability.
    ( doesn't like lots of requests at once)
    :param num_servers: The number of servers to create.
    :return:
    """
    batch_size = 10  # VMs to create at once
    with futures.ThreadPoolExecutor(batch_size) as pool:
        servers = [pool.submit(create_server) for _ in range(num_servers)]
    return [s.result() for s in futures.as_completed(servers)]


# Ensure that teardown will always occur at exit
register(teardown)
